@@ -1,6 +1,6 @@
 /*
      This file is part of libmicrohttpd
-     (C) 2007 Daniel Pittman and Christian Grothoff
+     (C) 2007, 2008 Daniel Pittman and Christian Grothoff
 
      This library is free software; you can redistribute it and/or
      modify it under the terms of the GNU Lesser General Public
@@ -50,7 +50,11 @@
  * Intentionally empty here to keep our memory footprint
  * minimal.
  */
+#if HAVE_MESSAGES
+#define REQUEST_TOO_BIG "<html><head><title>Request too big</title></head><body>Your HTTP header was too big for the memory constraints of this webserver.</body></html>"
+#else
 #define REQUEST_TOO_BIG ""
+#endif
 
 /**
  * Response text used when the request (http header) does not
@@ -59,7 +63,36 @@
  * Intentionally empty here to keep our memory footprint
  * minimal.
  */
+#if HAVE_MESSAGES
+#define REQUEST_LACKS_HOST "<html><head><title>&quot;Host:&quot; header required</title></head><body>In HTTP 1.1, requests must include a &quot;Host:&quot; header, and your HTTP 1.1 request lacked such a header.</body></html>"
+#else
 #define REQUEST_LACKS_HOST ""
+#endif
+
+/**
+ * Response text used when the request (http header) is
+ * malformed.
+ *
+ * Intentionally empty here to keep our memory footprint
+ * minimal.
+ */
+#if HAVE_MESSAGES
+#define REQUEST_MALFORMED "<html><head><title>Request malformed</title></head><body>Your HTTP request was syntactically incorrect.</body></html>"
+#else
+#define REQUEST_MALFORMED ""
+#endif
+
+/**
+ * Response text used when there is an internal server error.
+ *
+ * Intentionally empty here to keep our memory footprint
+ * minimal.
+ */
+#if HAVE_MESSAGES
+#define INTERNAL_ERROR "<html><head><title>Internal server error</title></head><body>Some programmer needs to study the manual more carefully.</body></html>"
+#else
+#define INTERNAL_ERROR ""
+#endif
 
 #define EXTRA_CHECKS MHD_YES
 
@@ -420,7 +453,7 @@ get_date_string (char *date, unsigned int max)
             days[now.tm_wday % 7],
             now.tm_mday,
             mons[now.tm_mon % 12],
-            now.tm_year, now.tm_hour, now.tm_min, now.tm_sec);
+            1900 + now.tm_year, now.tm_hour, now.tm_min, now.tm_sec);
 }
 
 /**
@@ -534,15 +567,15 @@ build_header_response (struct MHD_Connection *connection)
 }
 
 /**
- * We ran out of memory processing the
- * header.  Handle it properly by stopping to read data
- * and sending a HTTP 413 or HTTP 414 response.
+ * We encountered an error processing the request.
+ * Handle it properly by stopping to read data
+ * and sending the indicated response code and message.
  *
- * @param status_code the response code to send (413 or 414)
+ * @param status_code the response code to send (400, 413 or 414)
  */
 static void
-excessive_data_handler (struct MHD_Connection *connection,
-                        unsigned int status_code)
+transmit_error_response (struct MHD_Connection *connection,
+                         unsigned int status_code, const char *message)
 {
   struct MHD_Response *response;
 
@@ -551,10 +584,11 @@ excessive_data_handler (struct MHD_Connection *connection,
   connection->read_closed = MHD_YES;
 #if HAVE_MESSAGES
   MHD_DLOG (connection->daemon,
-            "Received excessively long header, closing connection.\n");
+            "Error %u (`%s') processing request, closing connection.\n",
+            status_code, message);
 #endif
-  response = MHD_create_response_from_data (strlen (REQUEST_TOO_BIG),
-                                            REQUEST_TOO_BIG, MHD_NO, MHD_NO);
+  response = MHD_create_response_from_data (strlen (message),
+                                            (void *) message, MHD_NO, MHD_NO);
   MHD_queue_response (connection, status_code, response);
   EXTRA_CHECK (connection->response != NULL);
   MHD_destroy_response (response);
@@ -632,10 +666,11 @@ MHD_connection_get_fdset (struct MHD_Connection *connection,
           if ((connection->read_buffer_offset == connection->read_buffer_size)
               && (MHD_NO == try_grow_read_buffer (connection)))
             {
-              excessive_data_handler (connection,
-                                      (connection->url != NULL)
-                                      ? MHD_HTTP_REQUEST_ENTITY_TOO_LARGE
-                                      : MHD_HTTP_REQUEST_URI_TOO_LONG);
+              transmit_error_response (connection,
+                                       (connection->url != NULL)
+                                       ? MHD_HTTP_REQUEST_ENTITY_TOO_LARGE
+                                       : MHD_HTTP_REQUEST_URI_TOO_LONG,
+                                       REQUEST_TOO_BIG);
               continue;
             }
           if (MHD_NO == connection->read_closed)
@@ -653,8 +688,31 @@ MHD_connection_get_fdset (struct MHD_Connection *connection,
           break;
         case MHD_CONNECTION_CONTINUE_SENT:
           if (connection->read_buffer_offset == connection->read_buffer_size)
-            try_grow_read_buffer (connection);
-          if (connection->read_buffer_offset < connection->read_buffer_size)
+            {
+              if ((MHD_YES != try_grow_read_buffer (connection)) &&
+                  (0 != (connection->daemon->options &
+                         (MHD_USE_SELECT_INTERNALLY |
+                          MHD_USE_THREAD_PER_CONNECTION))))
+                {
+                  /* failed to grow the read buffer, and the
+                     client which is supposed to handle the
+                     received data in a *blocking* fashion
+                     (in this mode) did not handle the data as
+                     it was supposed to!
+                     => we would either have to do busy-waiting
+                     (on the client, which would likely fail),
+                     or if we do nothing, we would just timeout
+                     on the connection (if a timeout is even
+                     set!).
+                     Solution: we kill the connection with an error */
+                  transmit_error_response (connection,
+                                           MHD_HTTP_INTERNAL_SERVER_ERROR,
+                                           INTERNAL_ERROR);
+                  continue;
+                }
+            }
+          if ((connection->read_buffer_offset < connection->read_buffer_size)
+              && (MHD_NO == connection->read_closed))
             do_fd_set (fd, read_fd_set, max_fd);
           break;
         case MHD_CONNECTION_BODY_RECEIVED:
@@ -747,10 +805,11 @@ get_next_header_line (struct MHD_Connection *connection)
                                       MHD_BUF_INC_SIZE);
           if (rbuf == NULL)
             {
-              excessive_data_handler (connection,
-                                      (connection->url != NULL)
-                                      ? MHD_HTTP_REQUEST_ENTITY_TOO_LARGE
-                                      : MHD_HTTP_REQUEST_URI_TOO_LONG);
+              transmit_error_response (connection,
+                                       (connection->url != NULL)
+                                       ? MHD_HTTP_REQUEST_ENTITY_TOO_LARGE
+                                       : MHD_HTTP_REQUEST_URI_TOO_LONG,
+                                       REQUEST_TOO_BIG);
             }
           else
             {
@@ -788,7 +847,8 @@ connection_add_header (struct MHD_Connection *connection,
       MHD_DLOG (connection->daemon,
                 "Not enough memory to allocate header record!\n");
 #endif
-      excessive_data_handler (connection, MHD_HTTP_REQUEST_ENTITY_TOO_LARGE);
+      transmit_error_response (connection, MHD_HTTP_REQUEST_ENTITY_TOO_LARGE,
+                               REQUEST_TOO_BIG);
       return MHD_NO;
     }
   hdr->next = connection->headers_received;
@@ -855,7 +915,8 @@ parse_cookie_header (struct MHD_Connection *connection)
 #if HAVE_MESSAGES
       MHD_DLOG (connection->daemon, "Not enough memory to parse cookies!\n");
 #endif
-      excessive_data_handler (connection, MHD_HTTP_REQUEST_ENTITY_TOO_LARGE);
+      transmit_error_response (connection, MHD_HTTP_REQUEST_ENTITY_TOO_LARGE,
+                               REQUEST_TOO_BIG);
       return MHD_NO;
     }
   memcpy (cpy, hdr, strlen (hdr) + 1);
@@ -1139,6 +1200,7 @@ do_read (struct MHD_Connection *connection)
     {
       /* other side closed connection */
       connection->read_closed = MHD_YES;
+      SHUTDOWN (connection->socket_fd, SHUT_RD);
       return MHD_NO;
     }
   connection->read_buffer_offset += bytes_read;
@@ -1150,7 +1212,7 @@ do_read (struct MHD_Connection *connection)
  * header (or footer).  Validate (check for ":") and prepare
  * to process.
  */
-static void
+static int
 process_header_line (struct MHD_Connection *connection, char *line)
 {
   char *colon;
@@ -1165,7 +1227,7 @@ process_header_line (struct MHD_Connection *connection, char *line)
                 "Received malformed line (no colon), closing connection.\n");
 #endif
       connection->state = MHD_CONNECTION_CLOSED;
-      return;
+      return MHD_NO;
     }
   /* zero-terminate header */
   colon[0] = '\0';
@@ -1179,6 +1241,7 @@ process_header_line (struct MHD_Connection *connection, char *line)
      with a space...) */
   connection->last = line;
   connection->colon = colon;
+  return MHD_YES;
 }
 
 /**
@@ -1188,8 +1251,9 @@ process_header_line (struct MHD_Connection *connection, char *line)
  * @param line the current input line
  * @param kind if the line is complete, add a header
  *        of the given kind
+ * @return MHD_YES if the line was processed successfully
  */
-static void
+static int
 process_broken_line (struct MHD_Connection *connection,
                      char *line, enum MHD_ValueKind kind)
 {
@@ -1207,26 +1271,37 @@ process_broken_line (struct MHD_Connection *connection,
                                   strlen (line) + strlen (last) + 1);
       if (last == NULL)
         {
-          excessive_data_handler (connection,
-                                  MHD_HTTP_REQUEST_ENTITY_TOO_LARGE);
-          return;
+          transmit_error_response (connection,
+                                   MHD_HTTP_REQUEST_ENTITY_TOO_LARGE,
+                                   REQUEST_TOO_BIG);
+          return MHD_NO;
         }
       tmp = line;
       while ((tmp[0] == ' ') || (tmp[0] == '\t'))
         tmp++;                  /* skip whitespace at start of 2nd line */
       strcat (last, tmp);
       connection->last = last;
-      return;                   /* possibly more than 2 lines... */
+      return MHD_YES;           /* possibly more than 2 lines... */
     }
-  if (MHD_NO == connection_add_header (connection,
-                                       last, connection->colon, kind))
+  EXTRA_CHECK ((last != NULL) && (connection->colon != NULL));
+  if ((MHD_NO == connection_add_header (connection,
+                                        last, connection->colon, kind)))
     {
-      excessive_data_handler (connection, MHD_HTTP_REQUEST_ENTITY_TOO_LARGE);
-      return;
+      transmit_error_response (connection, MHD_HTTP_REQUEST_ENTITY_TOO_LARGE,
+                               REQUEST_TOO_BIG);
+      return MHD_NO;
     }
   /* we still have the current line to deal with... */
   if (strlen (line) != 0)
-    process_header_line (connection, line);
+    {
+      if (MHD_NO == process_header_line (connection, line))
+        {
+          transmit_error_response (connection,
+                                   MHD_HTTP_BAD_REQUEST, REQUEST_MALFORMED);
+          return MHD_NO;
+        }
+    }
+  return MHD_YES;
 }
 
 /**
@@ -1594,7 +1669,16 @@ MHD_connection_handle_idle (struct MHD_Connection *connection)
         case MHD_CONNECTION_INIT:
           line = get_next_header_line (connection);
           if (line == NULL)
-            break;
+            {
+              if (connection->state != MHD_CONNECTION_INIT)
+                continue;
+              if (connection->read_closed)
+                {
+                  connection->state = MHD_CONNECTION_CLOSED;
+                  continue;
+                }
+              break;
+            }
           if (MHD_NO == parse_initial_message_line (connection, line))
             connection->state = MHD_CONNECTION_CLOSED;
           else
@@ -1603,20 +1687,46 @@ MHD_connection_handle_idle (struct MHD_Connection *connection)
         case MHD_CONNECTION_URL_RECEIVED:
           line = get_next_header_line (connection);
           if (line == NULL)
-            break;
+            {
+              if (connection->state != MHD_CONNECTION_URL_RECEIVED)
+                continue;
+              if (connection->read_closed)
+                {
+                  connection->state = MHD_CONNECTION_CLOSED;
+                  continue;
+                }
+              break;
+            }
           if (strlen (line) == 0)
             {
               connection->state = MHD_CONNECTION_HEADERS_RECEIVED;
               continue;
             }
-          process_header_line (connection, line);
+          if (MHD_NO == process_header_line (connection, line))
+            {
+              transmit_error_response (connection,
+                                       MHD_HTTP_BAD_REQUEST,
+                                       REQUEST_MALFORMED);
+              break;
+            }
           connection->state = MHD_CONNECTION_HEADER_PART_RECEIVED;
           continue;
         case MHD_CONNECTION_HEADER_PART_RECEIVED:
           line = get_next_header_line (connection);
           if (line == NULL)
-            break;
-          process_broken_line (connection, line, MHD_HEADER_KIND);
+            {
+              if (connection->state != MHD_CONNECTION_HEADER_PART_RECEIVED)
+                continue;
+              if (connection->read_closed)
+                {
+                  connection->state = MHD_CONNECTION_CLOSED;
+                  continue;
+                }
+              break;
+            }
+          if (MHD_NO ==
+              process_broken_line (connection, line, MHD_HEADER_KIND))
+            continue;
           if (strlen (line) == 0)
             {
               connection->state = MHD_CONNECTION_HEADERS_RECEIVED;
@@ -1631,6 +1741,8 @@ MHD_connection_handle_idle (struct MHD_Connection *connection)
           continue;
         case MHD_CONNECTION_HEADERS_PROCESSED:
           call_connection_handler (connection); /* first call */
+          if (connection->state == MHD_CONNECTION_CLOSED)
+            continue;
           if (need_100_continue (connection))
             {
               connection->state = MHD_CONNECTION_CONTINUE_SENDING;
@@ -1670,20 +1782,46 @@ MHD_connection_handle_idle (struct MHD_Connection *connection)
         case MHD_CONNECTION_BODY_RECEIVED:
           line = get_next_header_line (connection);
           if (line == NULL)
-            break;
+            {
+              if (connection->state != MHD_CONNECTION_BODY_RECEIVED)
+                continue;
+              if (connection->read_closed)
+                {
+                  connection->state = MHD_CONNECTION_CLOSED;
+                  continue;
+                }
+              break;
+            }
           if (strlen (line) == 0)
             {
               connection->state = MHD_CONNECTION_FOOTERS_RECEIVED;
               continue;
             }
-          process_header_line (connection, line);
+          if (MHD_NO == process_header_line (connection, line))
+            {
+              transmit_error_response (connection,
+                                       MHD_HTTP_BAD_REQUEST,
+                                       REQUEST_MALFORMED);
+              break;
+            }
           connection->state = MHD_CONNECTION_FOOTER_PART_RECEIVED;
           continue;
         case MHD_CONNECTION_FOOTER_PART_RECEIVED:
           line = get_next_header_line (connection);
           if (line == NULL)
-            break;
-          process_broken_line (connection, line, MHD_FOOTER_KIND);
+            {
+              if (connection->state != MHD_CONNECTION_FOOTER_PART_RECEIVED)
+                continue;
+              if (connection->read_closed)
+                {
+                  connection->state = MHD_CONNECTION_CLOSED;
+                  continue;
+                }
+              break;
+            }
+          if (MHD_NO ==
+              process_broken_line (connection, line, MHD_FOOTER_KIND))
+            continue;
           if (strlen (line) == 0)
             {
               connection->state = MHD_CONNECTION_FOOTERS_RECEIVED;
@@ -1795,6 +1933,7 @@ MHD_connection_handle_idle (struct MHD_Connection *connection)
               (connection->version == NULL) ||
               (0 != strcasecmp (MHD_HTTP_VERSION_1_1, connection->version)))
             {
+              /* http 1.0, version-less requests cannot be pipelined */
               connection->state = MHD_CONNECTION_CLOSED;
               MHD_pool_destroy (connection->pool);
               connection->pool = NULL;
@@ -1823,7 +1962,8 @@ MHD_connection_handle_idle (struct MHD_Connection *connection)
       break;
     }
   timeout = connection->daemon->connection_timeout;
-  if ((timeout != 0) && (time (NULL) - timeout > connection->last_activity))
+  if ((connection->socket_fd != -1) &&
+      (timeout != 0) && (time (NULL) - timeout > connection->last_activity))
     {
       connection_close_error (connection);
       return MHD_NO;
