@@ -1,6 +1,6 @@
 /*
      This file is part of libmicrohttpd
-     (C) 2006, 2007 Christian Grothoff (and other contributing authors)
+     (C) 2006, 2007, 2008 Christian Grothoff (and other contributing authors)
 
      This library is free software; you can redistribute it and/or
      modify it under the terms of the GNU Lesser General Public
@@ -84,7 +84,7 @@ extern "C"
 /**
  * Current version of the library.
  */
-#define MHD_VERSION 0x00000200
+#define MHD_VERSION 0x00030000
 
 /**
  * MHD-internal return codes.
@@ -333,6 +333,18 @@ enum MHD_OPTION
    */
   MHD_OPTION_NOTIFY_COMPLETED = 4,
 
+  /**
+   * Limit on the number of (concurrent) connections made to the
+   * server from the same IP address.  Can be used to prevent one
+   * IP from taking over all of the allowed connections.  If the
+   * same IP tries to establish more than the specified number of
+   * connections, they will be immediately rejected.  The option
+   * should be followed by an "unsigned int".  The default is
+   * zero, which means no limit on the number of connections
+   * from the same IP address.
+   */
+  MHD_OPTION_PER_IP_CONNECTION_LIMIT = 5,
+
 };
 
 /**
@@ -839,8 +851,13 @@ MHD_post_process (struct MHD_PostProcessor *pp,
 
 /**
  * Release PostProcessor resources.
+ * 
+ * @return MHD_YES if processing completed nicely,
+ *         MHD_NO if there were spurious characters / formatting
+ *                problems; it is common to ignore the return 
+ *                value of this function
  */
-void MHD_destroy_post_processor (struct MHD_PostProcessor *pp);
+int MHD_destroy_post_processor (struct MHD_PostProcessor *pp);
 
 
 #if 0                           /* keep Emacsens' auto-indent happy */
